'''
air septimanie (www.air-septimanie.com) Real-time Position System "OC-POS"

Copyright (c) 2019-now Markus Meurer, Saalfelden, Austria

This plugin serves as a replacement for the XAcars positioning system 
previously used by air septimanie, but which is no longer supported and available !
'''

from XPLMDefs import *
from XPLMProcessing import *
from XPLMDataAccess import *
from XPLMMenus import *

from XPLMUtilities import *

from XPWidgetDefs import *
from XPWidgets import *
from XPStandardWidgets import *

import random
import urllib2
import socket
import os
import datetime
import time
import threading

import sys

CALLBACK_INTERVAL = 5
# Default update interval = 15 seconds
DEFAULT_UPDATE_INTERVAL = 15
# Min and max values for update interval
MIN_UPDATE_INTERVAL = 15
MAX_UPDATE_INTERVAL = 120

# Other constants
FT_IN_M = 0.3048
M_S_IN_KTS = 1.943846
# -----------------------------------------------------------------------------------------------------------------------------------------------------
def getDateTimeAsUnixTimestamp(dtTime):
        return int(time.mktime(dtTime.timetuple()))
# -----------------------------------------------------------------------------------------------------------------------------------------------------
def getCurrentDateTimeAsUnixTimestamp():
        return getDateTimeAsUnixTimestamp(datetime.datetime.now())
# -----------------------------------------------------------------------------------------------------------------------------------------------------
def getDateTimeAsPerlString(dtTime):
        return dtTime.strftime("%d/%m/%Y|%H:%M")
# -----------------------------------------------------------------------------------------------------------------------------------------------------
def getCurrentDateTimeAsPerlString():
        return getDateTimeAsPerlString(datetime.datetime.now())
# -----------------------------------------------------------------------------------------------------------------------------------------------------
def getCurrentDateTimeAndTZOffsetInPerlFormat():
        return [getDateTimeAsPerlString(datetime.datetime.now()),0]
#---------------------------------------------------------------------------------------------------------------------------------------------
class ACARSUpdateThread(threading.Thread):
	OK = 0
	ERR_DEFAULT = -999
	ERR_OTHER_EXCEPTION = -976
	URL_ERROR = -1024
	# -----------------------------------------------------------------------------------------------------------------------------------
	def __init__(self):
                threading.Thread.__init__(self)
                self.updateFlightStatusCallback = 0
                self.asyncRequestedURL = ""
	# -----------------------------------------------------------------------------------------------------------------------------------
	def httpRequest(self,url):
		lErrCode = ACARSUpdateThread.OK
		try:
			response = urllib2.urlopen(url)
		except urllib2.HTTPError as e:
			lErrCode = -e.code
		except urllib2.URLError as e:
			lErrCode = ACARSUpdateThread.URL_ERROR
		except:
                        print "Unexpected error at: " + datetime.datetime.now().strftime("%H:%M:%S.%f")
                        print sys.exc_info()[0]
                        if (str(sys.exc_info()[0]).find("httplib.BadStatusLine") != -1):
                                lErrCode = ACARSUpdateThread.OK
                        else:
                                lErrCode = ACARSUpdateThread.ERR_OTHER_EXCEPTION
		else:
			html = response.read()
			lErrCode = int(html)

		return lErrCode
	# -----------------------------------------------------------------------------------------------------------------------------------
	def httpRequestAsync(self,url):
                lErrCode = self.httpRequest(url)
                if  (self.updateFlightStatusCallback != 0):
                        self.updateFlightStatusCallback(lErrCode)
                return ACARSUpdateThread.OK                        
	# -----------------------------------------------------------------------------------------------------------------------------------
	def run(self):
                self.httpRequestAsync(self.asyncRequestedURL)
#---------------------------------------------------------------------------------------------------------------------------------------------
class ACARSController:
	COL_PRIME_NUMBERS = [17, 29, 37, 11, 3, 13, 41, 47, 5, 31, 7, 19, 23, 43]
        LT = 0
        U = ""
        C1 = ""
        C2 = ""
        C3 = ""
        C4 = ""
        C5 = ""
	SIG = "[ASX-40]"

	FLYING_THRESHHOLD_FT = 10
	MIN_TAXI_THRESHHOLD = 7
	MIN_PB_THRESHHOLD = 2

	STATUS_UNDEFINED = 0
	STATUS_DEPARTURE = 1
	STATUS_PUSHBACK = 2
	STATUS_TAXI = 3
	STATUS_IN_FLIGHT = 4
	STATUS_LANDED = 5

	SERVER_STATUS_UNDEFINED = ""
	SERVER_STATUS_AUTOMATIC = "A"
	SERVER_STATUS_AT_AIRPORT_DEPARTURE = "D"
	SERVER_STATUS_AT_AIRPORT_ARRIVAL = "L"
	SERVER_SUBSTATUS_BLANK = ""
	SERVER_SUBSTATUS_DEPARTURE_PUSHBACK = "DP"
	SERVER_SUBSTATUS_DEPARTURE_TAXI = "DT"
	SERVER_SUBSTATUS_PAUSED = "AP"

	DEFAULT_SOCKET_TIMEOUT = 15

	OK = 0
	INI_FILE_NOT_FOUND = -990
	INI_FILE_INCORRECT =-991
	ERR_DEFAULT = -999
	URL_ERROR = -1024
	REQUEST_PENDING = -1025
	ERROR_THRESHHOLD = 3
	# -----------------------------------------------------------------------------------------------------------------------------------
	def __init__(self):
		self.acarsState = ACARSController.STATUS_UNDEFINED
		self.prevAcarsState = ACARSController.STATUS_UNDEFINED
		self.acarsServerStatus = ACARSController.SERVER_STATUS_UNDEFINED
		self.acarsServerSubStatus = ACARSController.SERVER_SUBSTATUS_BLANK
		self.isTrackingStarted = False
		self.pauseDurationInSeconds = -1
		self.prevPauseState = False
		self.pauseStartTime = 0
		self.uid = 0
		self.flightNumber = ""
		self.errorcount = 0
		self.lT = ACARSController.LT
		self.urlPrefix = self.uec(self.lT,ACARSController.U)
		self.cmdStartTracking = self.uec(self.lT,ACARSController.C1)
		self.cmdStatusUpdate = self.uec(self.lT,ACARSController.C2)
		self.cmdPositionUpdate = self.uec(self.lT,ACARSController.C3)
		self.cmdUpdateTakeOffInformation = self.uec(self.lT,ACARSController.C4)
		self.cmdUpdateLandingInformation = self.uec(self.lT,ACARSController.C5)
		self.INIFile = os.path.join('Resources', 'plugins','PythonScripts','PI_airSeptimanieRealtimePosition.ini')
                self.updateFlightStatusGUICallback = 0
                socket.setdefaulttimeout(ACARSController.DEFAULT_SOCKET_TIMEOUT)
	# -----------------------------------------------------------------------------------------------------------------------------------
	def reset(self):
		self.__init__()
	# -----------------------------------------------------------------------------------------------------------------------------------
	def resetStateMachine(self):
		self.acarsState = ACARSController.STATUS_UNDEFINED
		self.prevAcarsState = ACARSController.STATUS_UNDEFINED
		self.acarsServerStatus = ACARSController.SERVER_STATUS_UNDEFINED
		self.acarsServerSubStatus = ACARSController.SERVER_SUBSTATUS_BLANK
		self.isTrackingStarted = False
		self.flightNumber = ""
		self.errorcount = 0
	# -----------------------------------------------------------------------------------------------------------------------------------
	# add pause time and fuel left to arguments
	def flightStateMachine(self,agl,gs,er,ps,fl):
		self.prevAcarsState = self.acarsState
        	# -----------------------------------------------------------------------------------------------------------------------------------
		# Perform processing for pause time
		if (not self.prevPauseState and ps and (self.pauseStartTime == 0)):
        		self.pauseStartTime = int(time.mktime(datetime.datetime.now().timetuple()))
		elif (self.prevPauseState and not ps and (self.pauseStartTime != 0)):
                        if (self.pauseDurationInSeconds != -1):
        			self.pauseDurationInSeconds += (int(time.mktime(datetime.datetime.now().timetuple())) - self.pauseStartTime);
        		self.pauseStartTime = 0
		self.prevPauseState = ps
        	# -----------------------------------------------------------------------------------------------------------------------------------
		if (agl > ACARSController.FLYING_THRESHHOLD_FT):
			# Update takeoff information if permitted, i.e. if prev status was self.acarsServerStatus = ACARSController.SERVER_STATUS_AT_AIRPORT_DEPARTURE
			if (self.acarsServerStatus == ACARSController.SERVER_STATUS_AT_AIRPORT_DEPARTURE):
                                self.pauseDurationInSeconds = 0
                                self.pauseStartTime = 0
                                self.updateTakeOffInformation()
			self.acarsState = ACARSController.STATUS_IN_FLIGHT
			self.acarsServerStatus = ACARSController.SERVER_STATUS_AUTOMATIC
			self.acarsServerSubStatus = ACARSController.SERVER_SUBSTATUS_BLANK
		elif (self.acarsState == ACARSController.STATUS_IN_FLIGHT):
			self.acarsState = ACARSController.STATUS_LANDED
			self.acarsServerStatus = ACARSController.SERVER_STATUS_AT_AIRPORT_ARRIVAL
			self.acarsServerSubStatus = ACARSController.SERVER_SUBSTATUS_BLANK
			# Update landing information prior to stopping the tracking
			self.updateLandingInformation(max(int(self.pauseDurationInSeconds/60),0),fl)
			self.pauseDurationInSeconds = -1
                        self.stopTracking()
		elif ((self.acarsState == ACARSController.STATUS_DEPARTURE) or  (self.acarsState == ACARSController.STATUS_PUSHBACK) or (self.acarsState == ACARSController.STATUS_UNDEFINED)) and (gs >= ACARSController.MIN_TAXI_THRESHHOLD):
			self.acarsState = ACARSController.STATUS_TAXI
			self.acarsServerStatus = ACARSController.SERVER_STATUS_AT_AIRPORT_DEPARTURE 
			self.acarsServerSubStatus = ACARSController.SERVER_SUBSTATUS_DEPARTURE_TAXI 
		elif ((self.acarsState == ACARSController.STATUS_DEPARTURE) or (self.acarsState == ACARSController.STATUS_UNDEFINED)) and (gs >= ACARSController.MIN_PB_THRESHHOLD):
			self.acarsState = ACARSController.STATUS_PUSHBACK
			self.acarsServerStatus = ACARSController.SERVER_STATUS_AT_AIRPORT_DEPARTURE 
			self.acarsServerSubStatus = ACARSController.SERVER_SUBSTATUS_DEPARTURE_PUSHBACK 
		elif (self.acarsState == ACARSController.STATUS_UNDEFINED):
			self.acarsState = ACARSController.STATUS_DEPARTURE
			self.acarsServerStatus = ACARSController.SERVER_STATUS_AT_AIRPORT_DEPARTURE 
			self.acarsServerSubStatus = ACARSController.SERVER_SUBSTATUS_BLANK
	
		return self.acarsState
	# -----------------------------------------------------------------------------------------------------------------------------------
	def tp(self,l):
		sO = ""
		for lCounter in range(l):
			sO = sO + chr(65 + random.randint(0,25))

		return sO
	# -----------------------------------------------------------------------------------------------------------------------------------
	def ec(self,l, s1, s2, s3):
		sT = ""
		lS = len(ACARSController.COL_PRIME_NUMBERS)
		s4 = s3 + "|" + s2  + "|" + s1
	
		for lCounter in range(len(s4)):
			sT = sT + chr((ord(s4[lCounter]) * (1 + (l % ACARSController.COL_PRIME_NUMBERS[lCounter % lS])) % 26) + 65) + chr((int(ord(s4[lCounter]) * (1 + (l % ACARSController.COL_PRIME_NUMBERS[lCounter % lS]))/676)) + 65) + chr((int(ord(s4[lCounter]) * (1 + (l % ACARSController.COL_PRIME_NUMBERS[lCounter % lS]))/26) % 26) + 65)

		lCounter = 0
		sO = ""
		while (len(sT) > 0):
			lT = 2 + ((l % ACARSController.COL_PRIME_NUMBERS[lCounter % lS]) * ACARSController.COL_PRIME_NUMBERS[(lCounter + 1) % lS] % 3)
			sO = sO + sT[0:lT] + (self.tp(lT) if (lT <= len(sT)) else "")
			sT = sT[lT:len(sT)]
			lCounter = lCounter + 1
	
		return sO
	# -----------------------------------------------------------------------------------------------------------------------------------
	def uec(self,l,s):
                sR = ""
                sT = ""
                lCounter = 0
		lS = len(ACARSController.COL_PRIME_NUMBERS)

        	while (len(s) > 0):
			lT = 2 + ((l % ACARSController.COL_PRIME_NUMBERS[lCounter % lS]) * ACARSController.COL_PRIME_NUMBERS[(lCounter + 1) % lS] % 3)
                        sT = sT + (s[0:lT] if (lT < len(s)) else s)
                        s = s[2 * lT:1000000] if (lT < len(s)) else ""
			lCounter += 1

                lCounter = 0
                while  (lCounter < len(sT)):
                        sR = sR + chr(((ord(sT[lCounter+1:lCounter+2]) - 65) * 676 + (ord(sT[lCounter+2:lCounter+3]) - 65) * 26 + (ord(sT[lCounter:lCounter+1]) - 65)) / (l % ACARSController.COL_PRIME_NUMBERS[(lCounter/3) % lS] + 1))
                        lCounter += 3                        

                return sR
	# -----------------------------------------------------------------------------------------------------------------------------------
        def sendTakeOffInformationExt(self,l,fn,dt,tz):
                acarsUpdateThread = ACARSUpdateThread()
                acarsUpdateThread.updateFlightStatusCallback = self.updateFlightStatusCallback
                acarsUpdateThread.asyncRequestedURL = self.urlPrefix + self.cmdUpdateTakeOffInformation + "_" + str(l) + "_" + fn + "_" + dt + "_" + str(tz)
                acarsUpdateThread.start()
                return ACARSController.OK
	# -----------------------------------------------------------------------------------------------------------------------------------
        def updateTakeOffInformation(self):
                dt = getCurrentDateTimeAndTZOffsetInPerlFormat()
                return self.validateERR(self.sendTakeOffInformationExt(self.uid,self.flightNumber,dt[0],dt[1]))
	# -----------------------------------------------------------------------------------------------------------------------------------
        def sendLandingInformationExt(self,l,fn,dt,tz,pt,fl):
                acarsUpdateThread = ACARSUpdateThread()
                acarsUpdateThread.updateFlightStatusCallback = self.updateFlightStatusCallback
                acarsUpdateThread.asyncRequestedURL = self.urlPrefix + self.cmdUpdateLandingInformation + "_" + str(l) + "_" + fn + "_" + dt + "_" + str(tz) + "_" + str(pt) + "_" + ("%.0f" % fl)
                acarsUpdateThread.start()
                return ACARSController.OK
	# -----------------------------------------------------------------------------------------------------------------------------------
        def updateLandingInformation(self,pt,fl):
                dt = getCurrentDateTimeAndTZOffsetInPerlFormat()
                return self.validateERR(self.sendLandingInformationExt(self.uid,self.flightNumber,dt[0],dt[1],pt,fl))
	# -----------------------------------------------------------------------------------------------------------------------------------
	def requestStartTracking(self,l, s1, s2, s3,lt,ln,ag):
                acarsUpdateThread = ACARSUpdateThread()
		return acarsUpdateThread.httpRequest(self.urlPrefix + self.cmdStartTracking + "_" + str(l) + "_" + self.ec(l,s1,s2,s3) + "_" + ("%.6f" % lt) + "_" + ("%.6f" % ln) + "_" + ("%.0f" % ag))
	# -----------------------------------------------------------------------------------------------------------------------------------
	def sendGroundStatusExt(self,l,fn,lt,ln,el,gs,sq,hd):
                lErr = ACARSController.OK
		acarsUpdateThread = ACARSUpdateThread()
		acarsUpdateThread.updateFlightStatusCallback = self.updateFlightStatusCallback
                acarsUpdateThread.asyncRequestedURL = self.urlPrefix + self.cmdStatusUpdate + "_" + str(l) + "_" + fn + "_" + self.acarsServerStatus + "_" + self.acarsServerSubStatus + "_" + ("%.6f" % lt) + "_" + ("%.6f" % ln) + "_" + ("%.2f" % el) + "_" + ("%.2f" % hd) + "_" + ("%d" % sq) + "_" + ("%.0f" % gs)
                if (self.acarsState == ACARSController.STATUS_LANDED):
                        lErr = acarsUpdateThread.httpRequest(acarsUpdateThread.asyncRequestedURL)
                        self.updateFlightStatusCallback(lErr)
                else:
                        acarsUpdateThread.start()
                return lErr
	# -----------------------------------------------------------------------------------------------------------------------------------
	def sendInFlightStatusExt(self,l,fn,lt,ln,el,vs,gs,sq,hd,ps):
		acarsUpdateThread = ACARSUpdateThread()
		acarsUpdateThread.updateFlightStatusCallback = self.updateFlightStatusCallback
                acarsUpdateThread.asyncRequestedURL = self.urlPrefix + self.cmdPositionUpdate + "_" + str(l) + "_" + fn + "_" + ("%.6f" % lt) + "_" + ("%.6f" % ln) + "_" + ("%.2f" % el) + "_" + ("%.0f" % vs) + "_" + ("%.0f" % gs) + "_" + ("%d" % sq) + "_" + ("%.2f" % hd) + "_" + (ACARSController.SERVER_SUBSTATUS_PAUSED if (ps) else ACARSController.SERVER_SUBSTATUS_BLANK)
		acarsUpdateThread.start()
		return ACARSController.OK
	# -----------------------------------------------------------------------------------------------------------------------------------
	def startTracking(self,l,s1, s2, s3,lt,ln,ag):
                self.uid = self.requestStartTracking(l,s1, s2, s3,lt,ln,ag)
                if (self.uid > 0):
                        self.flightNumber = s3
                        self.isTrackingStarted = True
		
		return (self.isTrackingStarted, self.uid)
	# -----------------------------------------------------------------------------------------------------------------------------------
	def stopTracking(self):
                self.isTrackingStarted = False
	# -----------------------------------------------------------------------------------------------------------------------------------
	def validateERR(self,lErr):
                if (lErr == ACARSController.OK):
                        self.errorcount = 0
                else:
                        self.errorcount += 1

                if (self.errorcount > ACARSController.ERROR_THRESHHOLD):
                        self.isTrackingStarted = False
                        
		return lErr
	# -----------------------------------------------------------------------------------------------------------------------------------
	def sendGroundStatus(self,lt,ln,el,gs,sq,hd):
                return self.sendGroundStatusExt(self.uid,self.flightNumber,lt,ln,el,gs,sq,hd)
	# -----------------------------------------------------------------------------------------------------------------------------------
	def sendInFlightStatus(self,lt,ln,el,vs,gs,sq,hd,ps):
                return self.sendInFlightStatusExt(self.uid,self.flightNumber,lt,ln,el,vs,gs,sq,hd,ps)
	# -----------------------------------------------------------------------------------------------------------------------------------
	def updateFlightStatus(self,lt,ln,el,vs,gs,sq,hd,ps):
                lErr = ACARSController.OK

                if (self.acarsState == ACARSController.STATUS_IN_FLIGHT):
                        lErr = self.sendInFlightStatus(lt,ln,el,vs,gs,sq,hd,ps)
                else:
                        lErr = self.sendGroundStatus(lt,ln,el,gs,sq,hd)
                        
                return lErr
	# -----------------------------------------------------------------------------------------------------------------------------------
	def updateFlightStatusCallback(self,lErr):
                lErr = self.validateERR(lErr)
                if (self.updateFlightStatusGUICallback != 0):
                    self.updateFlightStatusGUICallback(lErr)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def readINIFile(self):
                lines = []
                lErrCode = ACARSController.OK
		if (os.path.exists(self.INIFile) and os.path.isfile(self.INIFile)):
			fHandle = open(self.INIFile, "r")
			lines = fHandle.read().splitlines()
			fHandle.close()

			if (len(lines) == 8):
                                if (lines[0] == ACARSController.SIG):
                                        self.lT = int(lines[1])
                                        self.urlPrefix = self.uec(self.lT,lines[2])
                                        self.cmdStartTracking = self.uec(self.lT,lines[3])
                                        self.cmdStatusUpdate = self.uec(self.lT,lines[4])
                                        self.cmdPositionUpdate = self.uec(self.lT,lines[5])
					self.cmdUpdateTakeOffInformation = self.uec(self.lT,lines[6])
					self.cmdUpdateLandingInformation = self.uec(self.lT,lines[7])
                                else:
                                        lErrCode = ACARSController.INI_FILE_INCORRECT
                        else:
                                lErrCode = ACARSController.INI_FILE_INCORRECT
                                        
                else:
                        lErrCode = ACARSController.INI_FILE_NOT_FOUND

                return lErrCode
	# -----------------------------------------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------------------------------------------------
class XPlaneDataController:
        def __init__(self):
                # Position and speed, ...
                self.geoLatitude = 0.0
                self.geoLongitude = 0.0
                self.elevationFeet = 0.0
                self.vsFPM = 0.0
                self.gsInKts = 0.0
                self.squawk = 0
                self.heading = 0.0
                self.paused = False
                self.enginesRunning = False
                self.aglFeet = 0
                self.fuelRemainingInKg = 0

                self.refGeoLatitude = 0
                self.refGeoLongitude = 0
                self.refElevationFeet = 0
                self.refVSFPM = 0
                self.refGSInKts = 0
                self.refSquawk = 0
                self.refHeading = 0
                self.refPaused = 0
                self.refAGLFeet = 0
                self.reffuelRemainingInKg = 0

                self.numberOfEngines = 0
                self.refNumberOfEngines = 0
                self.engineN1 = []
                self.refEngineN1 = 0
	# -----------------------------------------------------------------------------------------------------------------------------------
        def getPositionInformation(self):
                self.enginesRunning = False
                self.geoLatitude = self.getDataRefF('sim/flightmodel/position/latitude',self.refGeoLatitude,0)
                self.geoLongitude = self.getDataRefF('sim/flightmodel/position/longitude',self.refGeoLongitude,0)
                # Now use calibrated altitude - MM 4/1/2021
                # self.elevationFeet = self.getDataRefF('sim/flightmodel/position/elevation',self.refElevationFeet,0) / FT_IN_M
                self.elevationFeet = self.getDataRefF('sim/flightmodel/misc/h_ind',self.refElevationFeet,0) 
                self.vsFPM = self.getDataRefF('sim/flightmodel/position/vh_ind_fpm',self.refVSFPM,0)
                self.gsInKts = self.getDataRefF('sim/flightmodel/position/groundspeed',self.refGSInKts,0) * M_S_IN_KTS
                self.squawk = self.getDataRefI('sim/cockpit/radios/transponder_code',self.refSquawk,0)
                self.heading = self.getDataRefF('sim/flightmodel/position/psi',self.refHeading,0)
                self.paused = False if (self.getDataRefI('sim/time/paused',self.refPaused,0) == 0) else True
                self.aglFeet = self.getDataRefF('sim/flightmodel/position/y_agl',self.refAGLFeet,0) / FT_IN_M
                self.fuelRemainingInKg = self.getDataRefF('sim/flightmodel/weight/m_fuel_total',self.reffuelRemainingInKg,0)

                self.numberOfEngines = self.getDataRefI('sim/aircraft/engine/acf_num_engines',self.refNumberOfEngines,0)
                self.refEngineN1 = XPLMFindDataRef("sim/flightmodel/engine/ENGN_N1_")
                if (self.refEngineN1 != 0):
                        if (XPLMGetDatavf(self.refEngineN1,self.engineN1,0,self.numberOfEngines) != 0):
        			self.enginesRunning = (self.engineN1[0] >= 0.5) or (self.engineN1[1] >= 0.5)

                return (self.geoLatitude,self.geoLongitude,self.elevationFeet,self.vsFPM,self.gsInKts,self.squawk,self.heading,self.paused,self.enginesRunning,self.aglFeet,self.fuelRemainingInKg)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def getDataRefI(self,s,ref,dft):
                ret = dft
                
                if (ref == 0):
                        ref = XPLMFindDataRef(s)

                if (ref != 0):
                        ret = XPLMGetDatai(ref)

                return ret
	# -----------------------------------------------------------------------------------------------------------------------------------
        def getDataRefF(self,s,ref,dft):
                ret = dft
                
                if (ref == 0):
                        ref = XPLMFindDataRef(s)

                if (ref != 0):
                        ret = XPLMGetDataf(ref)

                return ret
#---------------------------------------------------------------------------------------------------------------------------------------------
class XPlaneGUIController:
        OK = 0
        INVALID_INT = -99999999999

        MSG_START_TRACKING = "Position Reporting started at %t"
        MSG_STOP_TRACKING = "Position Reporting stopped at %t"
        MSG_REPORT_SENT_OK = "Position Report successfully sent at %t"
        MSG_REPORT_SENT_ERR = "Error sending Position Report at %t (code %e)"
        MSG_REPORT_SENT_ERR_STOP = "Position Reporting stopped at %t due to multiple errors (code %e)"
        MSG_ERR_PEARL_OTHER = "Error obtaining flight information (code %e)"
        MSG_ERR_DICT = {
                                -1001: "Invalid flight number %f (must be 9Fnnnn, e.g. 9F7044)",
                                -1002: "The Report Interval must be betweem 15 and 600 seconds",
                                -1003: "Please specify your username, password and flightnumber",
                                -1004: "Your flight with flightnumber %f has already landed",
                                -1024: "Error sending data to air-septimanie.com at %t",
                                -1025: "Error - air-septimanie.com is busy processing your request",
                                -5410: "Invalid username, password or flight number (code %e)",
                                -5411: "Please check that you are really at the departure airport (code %e)",
                                -5412: "There are no active flights for your username '%u' on our database",
                                -5413: "There is no active flight with flight number %f for you",
                                -5414: "Your flight with flightnumber %f has already landed",
                                -6527: "No pilot does exist with username '%u'",
                                -6528: "Incorrect password for your username '%u'",
                                -6529: "Your pilot account has not yet been activated"
                        }
        
        def __init__(self):
                self.MainWindow = 0
                self.txtUserName = 0
                self.txtPassword = 0
                self.txtPositionReport = 0
                self.txtFlightNumber = 0
                self.lstStatus1 = 0
                self.lstStatus2 = 0
                self.cmdStartTracking = 0
                self.cmdStopTracking = 0
                self.cmdCloseWindow = 0
                self.StatusMessage = ""
	# -----------------------------------------------------------------------------------------------------------------------------------
        def createConfigurationWindow(self, x, y, w, h,updateInterval):
                x2 = x + w
                y2 = y - h

                # Create Window
                self.MainWindow = XPCreateWidget(x, y, x2, y2, 1, "OC-POS | Real-time Position System", 1,0 , xpWidgetClass_MainWindow)
                XPSetWidgetProperty(self.MainWindow, xpProperty_MainWindowHasCloseBoxes, 1)

                # Create top frame User/Password/Interval
                # Frame
                yRel = y - 25
                self.TopFrame = XPCreateWidget(x + 5, yRel, x2 - 5, yRel - 110, 1, "", 0, self.MainWindow, xpWidgetClass_SubWindow)
                XPSetWidgetProperty(self.TopFrame, xpProperty_SubWindowType, xpSubWindowStyle_SubWindow)
                # Username
                XPCreateWidget(x+12, yRel-9, x+110, yRel-25, 1, 'Your Username', 0, self.MainWindow, xpWidgetClass_Caption)
                self.txtUserName = XPCreateWidget(x+150, yRel-9, x2-15, yRel-26, 1, '', 0, self.MainWindow, xpWidgetClass_TextField)
                XPSetWidgetProperty(self.txtUserName, xpProperty_TextFieldType, xpTextEntryField)
                XPSetWidgetProperty(self.txtUserName, xpProperty_Enabled, 1)
                XPSetWidgetProperty(self.txtUserName, xpProperty_MaxCharacters, 60)

                # Password
                XPCreateWidget(x+12, yRel-34, x+110, yRel-50, 1, 'Your Password', 0, self.MainWindow, xpWidgetClass_Caption)
                self.txtPassword = XPCreateWidget(x+150, yRel-36, x2-15, yRel-51, 1, '', 0, self.MainWindow, xpWidgetClass_TextField)
                XPSetWidgetProperty(self.txtPassword, xpProperty_TextFieldType, xpTextEntryField)
                XPSetWidgetProperty(self.txtPassword, xpProperty_Enabled, 1)
                XPSetWidgetProperty(self.txtPassword, xpProperty_MaxCharacters, 60)
                XPSetWidgetProperty(self.txtPassword, xpProperty_PasswordMode, 1)

                # Flight Number
                XPCreateWidget(x+12, yRel-59, x+110, yRel-75, 1, 'Your Flight Number', 0, self.MainWindow, xpWidgetClass_Caption)
                self.txtFlightNumber = XPCreateWidget(x+150, yRel-61, x+200, yRel-76, 1, '', 0, self.MainWindow, xpWidgetClass_TextField)
                XPSetWidgetProperty(self.txtFlightNumber, xpProperty_Enabled, 1)
                XPSetWidgetProperty(self.txtFlightNumber, xpProperty_MaxCharacters, 6)

                # Position Report in seconds
                XPCreateWidget(x+12, yRel-84, x+145, yRel-100, 1, 'Position Report Interval', 0, self.MainWindow, xpWidgetClass_Caption)
                XPCreateWidget(x+185, yRel-84, x+270, yRel-100, 1, 'seconds', 0, self.MainWindow, xpWidgetClass_Caption)
                self.txtPositionReport = XPCreateWidget(x+150, yRel-86, x + 180, yRel-101, 1, str(updateInterval), 0, self.MainWindow, xpWidgetClass_TextField)
                XPSetWidgetProperty(self.txtPositionReport, xpProperty_TextFieldType, xpTextEntryField)
                XPSetWidgetProperty(self.txtPositionReport, xpProperty_Enabled, 1)
                XPSetWidgetProperty(self.txtPositionReport, xpProperty_MaxCharacters, 3)

                # Status Frame
                # Frame
                yRel = y - 140
                self.StatusFrame = XPCreateWidget(x + 5, yRel, x2 - 5, yRel - 75, 1, "", 0, self.MainWindow, xpWidgetClass_SubWindow)
                XPSetWidgetProperty(self.StatusFrame, xpProperty_SubWindowType, xpSubWindowStyle_SubWindow)
                # Status message list box
                XPCreateWidget(x+12, yRel-5, x+110, yRel-20, 1, 'Status', 0, self.MainWindow, xpWidgetClass_Caption)
                self.lstStatus1 = XPCreateWidget(x+10, yRel-22, x2-10, yRel-37,1,'', 0, self.MainWindow,xpWidgetClass_Caption)
                self.lstStatus2 = XPCreateWidget(x+10, yRel-42, x2-10, yRel-57,1,'', 0, self.MainWindow,xpWidgetClass_Caption)
 
                # Buttons
                yRel = y - 225
                # START TRACKING (initially visible and enabled)
                self.cmdStartTracking = XPCreateWidget(x2-240, yRel, x2-120, yRel-20,1, "START TRACKING", 0, self.MainWindow, xpWidgetClass_Button)
                XPSetWidgetProperty(self.cmdStartTracking, xpProperty_Enabled, 1)
        	XPSetWidgetProperty(self.cmdStartTracking, xpProperty_ButtonType, xpPushButton)

                # STOP TRACKING (initially invisible and enabled)
                self.cmdStopTracking = XPCreateWidget(x2-240, yRel, x2-120, yRel-20,0, "STOP TRACKING", 0, self.MainWindow, xpWidgetClass_Button)
                XPSetWidgetProperty(self.cmdStopTracking, xpProperty_Enabled, 1)
        	XPSetWidgetProperty(self.cmdStopTracking, xpProperty_ButtonType, xpPushButton)

                # START NEW FLIGHT (initially invisible and enabled)
                self.cmdStartNewFlight = XPCreateWidget(x2-255, yRel, x2-120, yRel-20,0, "START NEW FLIGHT", 0, self.MainWindow, xpWidgetClass_Button)
                XPSetWidgetProperty(self.cmdStartNewFlight, xpProperty_Enabled, 1)
        	XPSetWidgetProperty(self.cmdStartNewFlight, xpProperty_ButtonType, xpPushButton)

                # CLOSE WINDOW (always visible and enabled)
                self.cmdCloseWindow = XPCreateWidget(x2-110, yRel, x2-5, yRel-20,1, "CLOSE WINDOW", 0, self.MainWindow, xpWidgetClass_Button)
                XPSetWidgetProperty(self.cmdCloseWindow, xpProperty_Enabled, 1)
        	XPSetWidgetProperty(self.cmdCloseWindow, xpProperty_ButtonType, xpPushButton)

                # SET THE FIRST TABSTOP
        	XPSetKeyboardFocus(self.txtUserName)
 	# -----------------------------------------------------------------------------------------------------------------------------------
        def openConfigurationWindow(self,updateInterval):
                bIsNewWindow = False
                x = 300
                y = 700
                w = 330
                h = 253

                if (self.MainWindow == 0):
                        self.createConfigurationWindow(x, y, w, h,updateInterval)
                        bIsNewWindow = True
                elif not(XPIsWidgetVisible(self.MainWindow)):
			XPShowWidget(self.MainWindow)
			
                return bIsNewWindow
	# -----------------------------------------------------------------------------------------------------------------------------------
        def closeConfigurationWindow(self):
                if (XPIsWidgetVisible(self.MainWindow)):
			XPHideWidget(self.MainWindow)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def destroyConfigurationWindow(self):
                if (self.MainWindow != 0):
			XPDestroyWidget(self, self.MainWindow, 1)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def startNewFlight(self):
                XPSetKeyboardFocus(self.txtUserName)
                XPSetWidgetDescriptor(self.txtFlightNumber,"")
                self.resetStatus();
                XPShowWidget(self.cmdStartTracking)
                XPHideWidget(self.cmdStartNewFlight)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def startTracking(self,errCode,sUserName,sFlightNumber):
                if (errCode == XPlaneGUIController.OK):
                        sMessage = self.updateDynamicDataInMessage(XPlaneGUIController.MSG_START_TRACKING,XPlaneGUIController.OK,"","")
                        XPSetKeyboardFocus(self.cmdStopTracking)
                        XPSetWidgetProperty(self.txtUserName, xpProperty_Enabled, 0)
                        XPSetWidgetProperty(self.txtPassword, xpProperty_Enabled, 0)
                        XPSetWidgetProperty(self.txtFlightNumber, xpProperty_Enabled, 0)
                        XPSetWidgetProperty(self.txtPositionReport, xpProperty_Enabled, 0)
                        XPShowWidget(self.cmdStopTracking)
                        XPHideWidget(self.cmdStartTracking)
                        XPHideWidget(self.cmdStartNewFlight)
                else:
                        if (errCode == -5414):
                                XPHideWidget(self.cmdStopTracking)
                                XPHideWidget(self.cmdStartTracking)
                                XPShowWidget(self.cmdStartNewFlight)
                        
                        sMessage = self.getStartTrackingErrorMsg(errCode,sUserName,sFlightNumber)

                self.updateStatus(sMessage,False)  
	# -----------------------------------------------------------------------------------------------------------------------------------
        def stopTracking(self,errCode,acarsState):
                if (errCode == XPlaneGUIController.OK):
                        sMessage = self.updateDynamicDataInMessage(XPlaneGUIController.MSG_STOP_TRACKING,XPlaneGUIController.OK,"","")
                else:
                        sMessage = self.updateDynamicDataInMessage(XPlaneGUIController.MSG_REPORT_SENT_ERR_STOP,errCode,"","")

                self.updateStatus(sMessage,False)

                XPSetWidgetProperty(self.txtUserName, xpProperty_Enabled, 1)
                XPSetWidgetProperty(self.txtPassword, xpProperty_Enabled, 1)
                XPSetWidgetProperty(self.txtFlightNumber, xpProperty_Enabled, 1)
                XPSetWidgetProperty(self.txtPositionReport, xpProperty_Enabled, 1)
                        
                if (acarsState != ACARSController.STATUS_LANDED):
                        XPShowWidget(self.cmdStartTracking)
                        XPHideWidget(self.cmdStartNewFlight)
                else:
                        XPHideWidget(self.cmdStartTracking)
                        XPShowWidget(self.cmdStartNewFlight)
                                
                XPHideWidget(self.cmdStopTracking)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def resetStatus(self):
                self.updateStatus("",False)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def updateStatus(self,sMessage,bForceUpdate):
                self.StatusMessage = sMessage
                if (XPIsWidgetVisible(self.MainWindow) or bForceUpdate):
                        sMessage1 = ""
                        sMessage2 = ""
                        if (len(sMessage) > 45):
                                lPos = sMessage[0:45].rfind(" ")
                                if (lPos >= 0):
                                        sMessage1 = sMessage[0:lPos]
                                        sMessage2 = sMessage[lPos+1:100000]
                                else:
                                        sMessage1 = sMessage
                                        sMessage2 = ""
                        else:
                                sMessage1 = sMessage
                                sMessage2 = ""
                                
                        XPSetWidgetDescriptor(self.lstStatus1,sMessage1)
                        XPSetWidgetDescriptor(self.lstStatus2,sMessage2)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def windowHandler(self, inMessage, inWidget, inParam1, inParam2,acarsController):
                retValue = 0
                if ((inMessage == xpMsg_Create) and (inWidget == self.MainWindow) and (inParam1 == self.MainWindow)):
                        self.updateStatus(self.StatusMessage,True)
                elif ((inMessage == xpMsg_Shown) and (inWidget == self.MainWindow) and (inParam1 == self.MainWindow)):
                        self.updateStatus(self.StatusMessage,True)
                elif (inMessage == xpMessage_CloseButtonPushed):
                        self.closeConfigurationWindow()
                        retValue = 1
                elif (inMessage == xpMsg_PushButtonPressed):
                        if (inParam1 == self.cmdCloseWindow):
                                self.closeConfigurationWindow()
                                retValue = 1
                elif (inMessage == xpMsg_KeyPress):
                        key, flags, vkey = PI_GetKeyState(inParam1)
                        if (key == 9):
                                if (XPGetWidgetWithFocus() == self.txtUserName):
                                        XPSetKeyboardFocus(self.txtPositionReport if (flags & xplm_ShiftFlag) else self.txtPassword)
                                elif (XPGetWidgetWithFocus() == self.txtPassword):
                                        XPSetKeyboardFocus(self.txtUserName if (flags & xplm_ShiftFlag) else self.txtFlightNumber)
                                elif (XPGetWidgetWithFocus() == self.txtFlightNumber):
                                        XPSetKeyboardFocus(self.txtPassword if (flags & xplm_ShiftFlag) else self.txtPositionReport)
                                elif (XPGetWidgetWithFocus() == self.txtPositionReport):
                                        XPSetKeyboardFocus(self.txtFlightNumber if (flags & xplm_ShiftFlag) else self.txtUserName)
                                retValue = 1
                        elif (key == 27):
                                self.closeConfigurationWindow()
                                retValue = 1
                        
                return retValue
	# -----------------------------------------------------------------------------------------------------------------------------------
        def convertToInt(self, string):
                try: 
                    val = int(string)
                except ValueError:
                    val = XPlaneGUIController.INVALID_INT
                    
                return val
	# -----------------------------------------------------------------------------------------------------------------------------------
        def updateDynamicDataInMessage(self,sInput,errCode,sUsername,sFlightNumber):
                sMessage = sInput
                if (sMessage != ""):
                        sMessage = sMessage.replace("%e",str(errCode))
                        sMessage = sMessage.replace("%u",sUsername)
                        sMessage = sMessage.replace("%f",sFlightNumber)
                        sMessage = sMessage.replace("%t",datetime.datetime.now().strftime("%H:%M:%S"))

                return sMessage
	# -----------------------------------------------------------------------------------------------------------------------------------
        def updateFlightStatus(self,errCode,bIsStopped):
                sMessage = ""
                if (errCode == XPlaneGUIController.OK):
                        sMessage = XPlaneGUIController.MSG_REPORT_SENT_OK
                else:
                        if (errCode in XPlaneGUIController.MSG_ERR_DICT):
                                sMessage = XPlaneGUIController.MSG_ERR_DICT[errCode]
                                if (sMessage == ""):
                                        if (bIsStopped):
                                                sMessage = XPlaneGUIController.MSG_REPORT_SENT_ERR_STOP
                                        else:
                                                sMessage = XPlaneGUIController.MSG_REPORT_SENT_ERR
                                else:
                                        sMessage = XPlaneGUIController.MSG_ERR_PEARL_OTHER

                self.updateStatus(self.updateDynamicDataInMessage(sMessage,errCode,"",""),False)
 	# -----------------------------------------------------------------------------------------------------------------------------------
        def getStartTrackingErrorMsg(self,errCode,sUsername,sFlightNumber):
                if (errCode in XPlaneGUIController.MSG_ERR_DICT):
                        sErrMsg = XPlaneGUIController.MSG_ERR_DICT[errCode]
                        if ((sErrMsg == "") and (errCode != XPlaneGUIController.OK)):
                                sErrMsg = XPlaneGUIController.MSG_ERR_PEARL_OTHER
                else:
                        sErrMsg = XPlaneGUIController.MSG_ERR_PEARL_OTHER
                        
                return self.updateDynamicDataInMessage(sErrMsg,errCode,sUsername,sFlightNumber)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def preValidateInput(self,sUserName,sPassword,sFlightNumber,sPosInterval):
                errCode = XPlaneGUIController.OK
                sErrMsg = ""
                lPosInterval = self.convertToInt(sPosInterval)

                if ((sUserName == "") or (sPassword == "") or (sFlightNumber == "")):
                        errCode = -1003        
                else:
                        if ((lPosInterval >= 15) and (lPosInterval <= 600)):
                                if (((sFlightNumber[0:2] == "9F") or (sFlightNumber[0:2] == "9f")) and (self.convertToInt(sFlightNumber[2:len(sFlightNumber)-2]) != XPlaneGUIController.INVALID_INT) and (len(sFlightNumber) == 6)):
                                        pass
                                else:
                                        errCode = -1001  
                        else:
                                errCode = -1002  

                if (errCode != 0):
                        sErrMsg = self.getStartTrackingErrorMsg(errCode,"",sFlightNumber)

                return (errCode,sErrMsg)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def getDataFromWidget(self):
                sUserName = []
                sPassword = []
                sFlightNumber = []
                sPosInterval = []
                XPGetWidgetDescriptor(self.txtUserName, sUserName, 64)
                XPGetWidgetDescriptor(self.txtPassword, sPassword, 64)
                XPGetWidgetDescriptor(self.txtFlightNumber, sFlightNumber, 10)
                XPGetWidgetDescriptor(self.txtPositionReport, sPosInterval, 6)

                return (sUserName[0],sPassword[0],sFlightNumber[0].upper(),sPosInterval[0])
# ---------------------------------------------------------------------------------------------------------------------------------------------
class PythonInterface:
        
	def XPluginStart(self):
		self.Name = "OC-POS Real-time Positioning"
		self.Sig =  "airSeptimaniePositioning.airseptimanie.MM"
		self.Desc = "Real-time position reporting system for air septimanie flights (www.air-septimanie.com)"
		self.VERSION="v1.0"

		self.callbackInterval = CALLBACK_INTERVAL
		self.idParent = 0
		self.callbackCounter = 0
		self.updateInterval = DEFAULT_UPDATE_INTERVAL
		
		self.callback = self.flightLoopCallback
		self.menuHandlerCallback = self.menuHandler

                self.myXPlaneDataController = XPlaneDataController()

                self.myAcarsController = ACARSController()
                self.myAcarsController.updateFlightStatusGUICallback = self.updateFlightStatusCallback
                self.myAcarsController.readINIFile()

                self.myGUIController = XPlaneGUIController()

                self.constructMenu()
                XPLMRegisterFlightLoopCallback(self, self.callback, 0, 0)
		
		return self.Name, self.Sig, self.Desc
	# -----------------------------------------------------------------------------------------------------------------------------------
	def XPluginStop(self):
		XPLMUnregisterFlightLoopCallback(self, self.callback, 0)
                self.myGUIController.destroyConfigurationWindow()
		if (self.idParent != 0):
                        XPLMDestroyMenu(self, self.idParent)
                pass
 	# -----------------------------------------------------------------------------------------------------------------------------------
	def XPluginEnable(self):
		return 1
	# -----------------------------------------------------------------------------------------------------------------------------------
	def XPluginDisable(self):
		pass
	# -----------------------------------------------------------------------------------------------------------------------------------
	def XPluginReceiveMessage(self, inFromWho, inMessage, inParam):
		pass
	# -----------------------------------------------------------------------------------------------------------------------------------
	def menuHandler(self, inMenuRef, inItemRef):
        	if (inItemRef == 102):
                        if (self.myGUIController.openConfigurationWindow(self.updateInterval)):
                                self.mainWindowHandlerCB = self.mainWindowHandler
                                XPAddWidgetCallback(self, self.myGUIController.MainWindow, self.mainWindowHandlerCB)
	# -----------------------------------------------------------------------------------------------------------------------------------
	def constructMenu(self):
		self.mnuItemParent = XPLMAppendMenuItem(XPLMFindPluginsMenu(), self.Name, 0, 1)
		self.idParent = XPLMCreateMenu(self, self.Name , XPLMFindPluginsMenu(), self.mnuItemParent, self.menuHandlerCallback, 0)
		self.mnuItemShowWindow = XPLMAppendMenuItem(self.idParent, "Show Configuration Window", 102, 1)
                return 0
	# -----------------------------------------------------------------------------------------------------------------------------------
        def flightLoopCallback(self, elapsedMe, elapsedSim, counter, refcon):
                lErr = ACARSController.OK
                if (self.myAcarsController.isTrackingStarted):
                        self.myXPlaneDataController.getPositionInformation()
                        self.myAcarsController.flightStateMachine(self.myXPlaneDataController.aglFeet,self.myXPlaneDataController.gsInKts,self.myXPlaneDataController.enginesRunning,self.myXPlaneDataController.paused,self.myXPlaneDataController.fuelRemainingInKg)
                        if (self.myAcarsController.acarsState != self.myAcarsController.prevAcarsState):
                                lErr = self.myAcarsController.updateFlightStatus(self.myXPlaneDataController.geoLatitude,self.myXPlaneDataController.geoLongitude,self.myXPlaneDataController.elevationFeet,self.myXPlaneDataController.vsFPM,self.myXPlaneDataController.gsInKts,self.myXPlaneDataController.squawk,self.myXPlaneDataController.heading,self.myXPlaneDataController.paused)
                        elif (((self.callbackCounter * self.callbackInterval) % self.updateInterval) < self.callbackInterval):
                                lErr = self.myAcarsController.updateFlightStatus(self.myXPlaneDataController.geoLatitude,self.myXPlaneDataController.geoLongitude,self.myXPlaneDataController.elevationFeet,self.myXPlaneDataController.vsFPM,self.myXPlaneDataController.gsInKts,self.myXPlaneDataController.squawk,self.myXPlaneDataController.heading,self.myXPlaneDataController.paused)

                        if ((lErr != ACARSController.OK) and (lErr is not None)):
                                self.updateFlightStatusCallback(lErr)

                self.callbackCounter += 1
                if (not self.myAcarsController.isTrackingStarted):
                        return 0
                else:
                        return self.callbackInterval
	# -----------------------------------------------------------------------------------------------------------------------------------
	def updateFlightStatusCallback(self, lErr):
                if lErr is None:
                        lErr = XPlaneGUIController.OK
                if (not self.myAcarsController.isTrackingStarted):
                        self.myGUIController.stopTracking(lErr,self.myAcarsController.acarsState)
                        XPLMSetFlightLoopCallbackInterval(self,self.callback, 0, 0, 0)
                else:
                        self.myGUIController.updateFlightStatus(lErr,not self.myAcarsController.isTrackingStarted)
	# -----------------------------------------------------------------------------------------------------------------------------------
        def mainWindowHandler(self, inMessage, inWidget, inParam1, inParam2):
                retValue  = self.myGUIController.windowHandler(inMessage, inWidget, inParam1, inParam2,self.myAcarsController)
                if ((retValue == 0) and (inMessage == xpMsg_PushButtonPressed)):
                        if (inParam1 == self.myGUIController.cmdStartTracking):
                                self.startTracking()
                        elif (inParam1 == self.myGUIController.cmdStopTracking):
                                self.stopTracking()
                        elif (inParam1 == self.myGUIController.cmdStartNewFlight):
                                self.startNewFlight()
                        else:
                                retValue = 0
                elif ((retValue == 0) and (inMessage == xpMsg_KeyPress)):
                        key, flags, vkey = PI_GetKeyState(inParam1)
                        if (key == 13):
                                if (XPIsWidgetVisible(self.myGUIController.cmdStartTracking)):
                                        self.startTracking()
                                elif (XPIsWidgetVisible(self.myGUIController.cmdStopTracking)):
                                        self.stopTracking()
                                elif (XPIsWidgetVisible(self.myGUIController.cmdStartNewFlight)):
                                        self.startNewFlight()

                                retValue = 1;     
                else:
                        retValue = 0
                        
                return retValue
	# -----------------------------------------------------------------------------------------------------------------------------------
	def startTracking(self):
                (sUserName,sPassword,sFlightNumber,sPosInterval) = self.myGUIController.getDataFromWidget()
                (errCode,sErrMsg) = self.myGUIController.preValidateInput(sUserName,sPassword,sFlightNumber,sPosInterval)
                if (errCode == XPlaneGUIController.OK):
                        self.myXPlaneDataController.getPositionInformation()
                        (bStarted,uid) = self.myAcarsController.startTracking(self.myAcarsController.lT,sUserName,sPassword,sFlightNumber,self.myXPlaneDataController.geoLatitude,self.myXPlaneDataController.geoLongitude,self.myXPlaneDataController.aglFeet)
                        if (bStarted):
                                self.updateInterval = int(sPosInterval)
                                self.myGUIController.startTracking(XPlaneGUIController.OK,sUserName,sFlightNumber)
                                XPLMSetFlightLoopCallbackInterval(self,self.callback, self.updateInterval, 0, 0)
                        else:
                                self.myGUIController.startTracking(uid,sUserName,sFlightNumber)
                else:
                        self.myGUIController.updateStatus(sErrMsg,False)                               
	# -----------------------------------------------------------------------------------------------------------------------------------
	def stopTracking(self):
                XPLMSetFlightLoopCallbackInterval(self, self.callback, 0, 0, 0)
                self.myAcarsController.stopTracking()
                self.myGUIController.stopTracking(XPlaneGUIController.OK,self.myAcarsController.acarsState)
	# -----------------------------------------------------------------------------------------------------------------------------------
	def startNewFlight(self):
                self.myAcarsController.resetStateMachine()
                self.myGUIController.startNewFlight()
	# -----------------------------------------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------------------------------------------------
